'''

 Filter to implement the stretching of each colour in the INPUT IMAGE
 (like Gimp interactive Auto White Balance) plus complementary procedures
 for removing the remaining colour cast to ensure a white balancing effect 
 
 .........................................................................
  Prerequisites:
 - none (uses std Gimp environment)
 .........................................................................
 
 Release 2.0   source from v.1 then added color cast detection thru histogram means
               (released as WhiteBalanceStretch2 1.0)
 Release 2.A   fixed a bug (flatten was wrong, merge_down has to be used) (internal)
 Release 2.B   A + used mid value instead of avg + introduced a percent of application (internal)
 Release 2.C   B + added pixelize, invert and overlay (internal)
 Release 2.D   C + added user's choise between Full or Moderate effect
               (released as WhiteBalanceStretch2 1.1)
 Release 2.1   MrQ: update for Gimp-2.10, added work on a copy.
 
 Flow implemented to get the final result:       
 
 1. Verify the requested percentage of erratic dark values in Red, Green, Blue
    (limit = 0.05%) to exclude
 2. Verify the requested percentage of erratic bright values in Red, Green, Blue
    (limit = 0.05%) to exclude
 3. Apply the Levels adjustments using min-max previously detected
 4. Calculate Color cast still present after the previous procedure using histograms
 5. Modify colours accordingly, either at 100% (full option) or at 66% (moderate option)
 6. If moderate option active, duplicate use pixelize + invert + LAYER_MODE_SOFTLIGHT_LEGACY at 33%
 
 ----------------------------------------------------------------
 COPYRIGHT NOTICE
 ----------------------------------------------------------------
 
 This program is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 3 of the License, or
 (at your option) any later version.  
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program if not, you can view the GNU General Public
 License version 3 at the web site http://www.gnu.org/licenses/gpl-3.0.html
 Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
 Ave, Cambridge, MA 02139, USA.
 
 -------------------------------------------------------------------
 other info
 -------------------------------------------------------------------
 
'''
#!/usr/bin/env python
from gimpfu import *

# ===================
# ANCILLARY FUNCTION
# ===================

def DeCast (img, layer, percR, percG, percB, fact) :    # 2.2
    
    # Reduces the cast in a layer, using the 3 calculated percentages of cast.
    # Quite fast, since it makes use of tiles.
    #
    # Indicates that the process has started.
    gimp.progress_init("Reducing cast tone in " + layer.name + "...")

    # Get the layer position.
    pos = 0;
    for i in range(len(img.layers)):
        if(img.layers[i] == layer):
            pos = i
    
    # Create a new layer to save the results (otherwise is not possible to undo the operation).
    newLayer = gimp.Layer(img, layer.name + " temp", layer.width, layer.height, layer.type, layer.opacity, layer.mode)
    img.add_layer(newLayer, pos)
    layerName = layer.name
    
    # Clear the new layer.
    pdb.gimp_edit_clear(newLayer)
    newLayer.flush()
    
    # Convert the pixels to gray scale.
    try:
        # Calculate the number of tiles.
		
		# MrQ replace 64 to 128
        tn = int(layer.width / 128)
        if(layer.width % 128 > 0):
            tn += 1
        tm = int(layer.height / 128)
        if(layer.height % 128 > 0):

            tm += 1
        # Iterate over the tiles.
        for i in range(tn):
            for j in range(tm):
                # Update the progress bar.
                gimp.progress_update(float(i*tm + j) / float(tn*tm))
                #gimp.message("before getting a tile")
                # Get the tiles.
                srcTile = layer.get_tile(False, j, i)
                dstTile = newLayer.get_tile(False, j, i)
        
                # Iterate over the pixels of each tile.
                for x in range(srcTile.ewidth):
                    for y in range(srcTile.eheight):
                        # Get the pixel and calculate his gray value.
                        pixel = srcTile[x,y]
                        red = ord(pixel[0])
                        green = ord(pixel[1])
                        blue = ord(pixel[2])
                        gray = (ord(pixel[0]) + ord(pixel[1]) + ord(pixel[2]))/3
                         
                        newred = min(int(red/percR),255)
                        newgreen = min(int(green/percG),255)
                        newblue = min(int(blue/percB),255)
                        #gimp.message("new values "+str(newred)+","+str(newgreen)+","+str(newblue))
                        # -------------- r.2.2 ----------------------
                        newred = int((newred*fact)+(red*(1.0-fact)))
                        newgreen = int((newgreen*fact)+(green*(1.0-fact)))
                        newblue = int((newblue*fact)+(blue*(1.0-fact)))
                        # -------------- r.2.2 ----------------------
                                     
                        #res = chr(gray) + chr(gray) + chr(gray)
                        res = chr(newred) + chr(newgreen) + chr(newblue)
                        
                        # If the image has an alpha channel (or any other channel) copy his values.
                        if(len(pixel) > 3):
                            for k in range(len(pixel)-3):
                                res += pixel[k+3]
                                
                        # Save the value in the result layer.
                        dstTile[x,y] = res
        
        # Update the new layer.
        newLayer.flush()
        newLayer.merge_shadow(True)
        newLayer.update(0, 0, newLayer.width, newLayer.height)
        
        # Remove the old layer.
        img.remove_layer(layer)
        
        # Change the name of the new layer (two layers can not have the same name).
        newLayer.name = layerName
    except Exception as err:
        gimp.message ("Unexpected error: " + str(err))
    
    # End progress.
    pdb.gimp_progress_end()
    
    return(newLayer)


#==================================================================================================
# MAIN MODULE
#--------------------------------------------------------------------------------------------------

def WhiteBalanceStretch2 (inImage, inDrawable, addColour, Strength) :

    ##### MrQ work on a copy
    active=pdb.gimp_image_get_active_layer(inImage)	
    pdb.gimp_item_set_visible(active, 1)	
    temp_layer=pdb.gimp_layer_copy(active,1)
    pdb.gimp_image_insert_layer(inImage, temp_layer, None, -1)
    pdb.gimp_image_raise_layer_to_top(inImage, temp_layer)
    inDrawable=temp_layer
	#####
	
    pdb.gimp_context_push
    currFG = pdb.gimp_context_get_foreground()
    currBG = pdb.gimp_context_get_background()
    
    # Set up an undo group, so the operation will be undone in one step.
    pdb.gimp_image_undo_group_start(inImage)
    
    name = inDrawable.name
    if '.' in name:
        split = name.split(".", 1)
        name = split[0]
    width = inDrawable.width
    height = inDrawable.height
    
    # =============================
    # REMOVE ERRATIC EXTREME VALUES
    # =============================
    
    erraticPercent = 0.05
    nonerratic = 0.10
    
    # RED 

    # =========== 
    # ZERO QUOTE
    # ===========  
    
    fromV = 0
    toV = 1
    percent = 0
    while (percent < erraticPercent) :  
        (mean, std_dev, median, pixels, count, percentile) = pdb.gimp_histogram (inDrawable, HISTOGRAM_RED, fromV, toV)
        percent = percentile * 100
        toV += 1
    #gimp.message ("percent dark RED erratic = " + str(percent))
    if percent < nonerratic :
        MinRED = toV - 1
    else :
        MinRED = max (0, toV-2)

    # =========== 
    # 255 QUOTE
    # ===========  
    
    fromV = 254
    toV = 255
    percent = 0
    while (percent < erraticPercent) :  
        (mean, std_dev, median, pixels, count, percentile) = pdb.gimp_histogram (inDrawable, HISTOGRAM_RED, fromV, toV)
        percent = percentile * 100  
        fromV -= 1
    #gimp.message ("percent bright RED errratic = " + str(percent))
    if percent < nonerratic :
        MaxRED = fromV + 1
    else :
        MaxRED = min (255, fromV+2)
    
    # GREEN 

    # =========== 
    # ZERO QUOTE
    # ===========  
    
    fromV = 0
    toV = 1
    percent = 0
    while (percent < erraticPercent) :  
        (mean, std_dev, median, pixels, count, percentile) = pdb.gimp_histogram (inDrawable, HISTOGRAM_GREEN, fromV, toV)
        percent = percentile * 100  
        toV += 1
    #gimp.message ("percent dark GREEN erratic = " + str(percent))
    if percent < nonerratic :
        MinGREEN = toV - 1
    else :
        MinGREEN = max (0, toV-2)

    # =========== 
    # 255 QUOTE
    # ===========  
    
    fromV = 254
    toV = 255
    percent = 0
    while (percent < erraticPercent) :  
        (mean, std_dev, median, pixels, count, percentile) = pdb.gimp_histogram (inDrawable, HISTOGRAM_GREEN, fromV, toV)
        percent = percentile * 100  
        fromV -= 1
    #gimp.message ("percent bright GREEN erratic = " + str(percent))
    if percent < nonerratic :
        MaxGREEN = fromV + 1
    else :
        MaxGREEN = min (255, fromV+2)
    
    # BLUE 

    # =========== 
    # ZERO QUOTE
    # ===========  
    
    fromV = 0
    toV = 1
    percent = 0
    while (percent < erraticPercent) :  
        (mean, std_dev, median, pixels, count, percentile) = pdb.gimp_histogram (inDrawable, HISTOGRAM_BLUE, fromV, toV)
        percent = percentile * 100  
        toV += 1
    #gimp.message ("percent dark BLUE erratic = " + str(percent))
    if percent < nonerratic :
        MinBLUE = toV - 1
    else :
        MinBLUE = max (0, toV-2)

    # =========== 
    # 255 QUOTE
    # ===========  
    
    fromV = 254
    toV = 255
    percent = 0
    while (percent < erraticPercent) :  
        (mean, std_dev, median, pixels, count, percentile) = pdb.gimp_histogram (inDrawable, HISTOGRAM_BLUE, fromV, toV)
        percent = percentile * 100  
        fromV -= 1
    #gimp.message ("percent bright BLUE erratic = " + str(percent))
    if percent < nonerratic :
        MaxBLUE = fromV + 1
    else :
        MaxBLUE = min (255, fromV+2)
    
    #gimp.message ("RED range = "+str(MinRED)+"-"+str(MaxRED))
    #gimp.message ("GREEN range = "+str(MinGREEN)+"-"+str(MaxGREEN))
    #gimp.message ("BLUE range = "+str(MinBLUE)+"-"+str(MaxBLUE))
    
    # =============================
    # APPLY MODIFICATIONS TO LEVELS
    # =============================
    
    pdb.gimp_levels (inDrawable, 1, MinRED, MaxRED, 1, 0, 255)
    pdb.gimp_levels (inDrawable, 2, MinGREEN, MaxGREEN, 1, 0, 255)
    pdb.gimp_levels (inDrawable, 3, MinBLUE, MaxBLUE, 1, 0, 255)
    
    # ===========================================
    # VERSION 2.0 Additions: remove residual cast 
    # ===========================================
    
    # Calculate the residual cast
    
    (meanR, stddevR, medianR, pixelsR, countR, percentR) = pdb.gimp_histogram(inDrawable, 1, 0, 255)    # RED
    (meanG, stddevG, medianG, pixelsG, countG, percentG) = pdb.gimp_histogram(inDrawable, 2, 0, 255)    # GREEN
    (meanB, stddevB, medianB, pixelsB, countB, percentB) = pdb.gimp_histogram(inDrawable, 3, 0, 255)    # BLUE
    #gimp.message ("values = "+str(meanR)+","+str(meanG)+","+str(meanB))
    maxmean = max(meanR, meanG, meanB)
    minmean = min(meanR, meanG, meanB)
    if Strength == 0:   # moderate
        meangray = int(meanR+meanG+meanB-maxmean-minmean)   # take the mid value
    else:               # full
        meangray = int((meanR+meanG+meanB)/3)               # take the average
    
    #gimp.message ("mean values = "+str(meanR)+","+str(meanG)+","+str(meanB))
    castR = float((meanR*1.000)/(meangray*1.000))
    castG = float((meanG*1.000)/(meangray*1.000))
    castB = float((meanB*1.000)/(meangray*1.000))
    #gimp.message ("cast values % = "+str(castR)+","+str(castG)+","+str(castB))
    if Strength == 0:   # moderate
        howmuch = 0.66   
    else:               # full
        howmuch = 1.00
    inDrawable = DeCast (inImage, inDrawable, castR, castG, castB, howmuch)      
    #gimp.message ("decasted")
    
    if addColour == True:
        duplicate = inDrawable.copy()
        inImage.add_layer (duplicate, 0)
        duplicate.name = "Duplicated"
        duplicate.mode = LAYER_MODE_SOFTLIGHT_LEGACY
        duplicate.opacity = 90.0
        pdb.gimp_hue_saturation (duplicate, 0, 0, 0, +33)
        pdb.gimp_desaturate (duplicate)
        invert = duplicate.copy()
        invert.name = "Inverted"
        inImage.add_layer (invert, 0)
        invert.mode = LAYER_MODE_SOFTLIGHT_LEGACY
        invert.opacity = 45.0
        pdb.plug_in_vinvert (inImage, invert)
        pdb.plug_in_gauss_iir2 (inImage, invert, width/100, height/100)
        #  it was : pdb.gimp_image_flatten (inImage)  # bug, fixed below  
        inDrawable = pdb.gimp_image_merge_down (inImage, duplicate, CLIP_TO_BOTTOM_LAYER)
        inDrawable = pdb.gimp_image_merge_down (inImage, invert, CLIP_TO_BOTTOM_LAYER)
   
    # Added 
    if Strength == 0:   # moderate
        pixel = inDrawable.copy()
        inImage.add_layer (pixel, 0)
        pixel.name = "Pixelated"
        pixel.mode = LAYER_MODE_OVERLAY_LEGACY
        pixel.opacity = 33.0
        pdb.plug_in_pixelize2 (inImage, pixel, width, height)
        pdb.gimp_invert (pixel)
        inDrawable = pdb.gimp_image_merge_down (inImage, pixel, CLIP_TO_BOTTOM_LAYER)
    
    # Close the undo group.
    
    pdb.gimp_image_undo_group_end(inImage)
    
    pdb.gimp_context_pop
    pdb.gimp_context_set_foreground(currFG)
    pdb.gimp_context_set_background(currBG)

    if Strength == 0:   # moderate
        inDrawable.name = name+" auto white balanced (moderate)"
    else:               # full
        inDrawable.name = name+" auto white balanced (full)"

    pdb.gimp_displays_flush
    
    return(inDrawable)

# This is the plugin registration function
register(
    "WhiteBalanceStretch2",    
    "Perform a white balance stretch of a drawable and removes residual cast (release 2.1).",   
    "Perform a white balance stretch of a drawable and removes residual cast (release 2.1).",
    "Diego", 
    "Diego Nassetti ", 
    "December 2016",
    "White Balance Stretch2", 
    "RGB*", 
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "drawable", "Input drawable", None),
      (PF_TOGGLE, "enhance", "Enhance outcome (post-process)?", True),
      (PF_OPTION, "strength", "Strength of colours de-casting", 0,
                                                                ("Moderate",                                                          
                                                                "Full",
                                                                )),
    ], 
    [
      (PF_DRAWABLE, "outdrawable", "Output drawable", None),
    ],
    WhiteBalanceStretch2,
    menu="<Image>/Diego/Colours",
    )

main()
